# -*- mode: python; coding: utf-8 -*-
# Copyright 2013-2021 Chris Beaumont and the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""Computations for the TOAST projection scheme and tile pyramid format.

For all TOAST maps, the north pole is in the dead center of the virtual image
square, the south pole is at all four of its corners, and the equator is a
diamond connecting the midpoints of the four sides of the square. (See Figure 3
of McGlynn+ 2019, DOI:10.3847/1538-4365/aaf79e).

For TOAST maps of the sky, the line of RA (lon) = 0 in the northern hemisphere
extends from the center of the square to the right, as in the Figure 3 mentioned
above. The RA = 90 line goes from the center up, and so on counter-clockwise
around the square.

For TOAST planetary maps, the lon = 0 line in the northern hemisphere extends
from the center of the square to the *left*. The lon = 90 line extends
downwards, and increasing longitude results in counter-clockwise motion around
the square as for sky maps. In other words, the longitudinal orientation is
rotated by 180 degrees.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
count_tiles_matching_filter
create_single_tile
generate_tiles
generate_tiles_filtered
sample_layer
sample_layer_filtered
Tile
ToastCoordinateSystem
toast_pixel_for_point
toast_tile_area
toast_tile_for_point
toast_tile_get_coords
'''.split()

from collections import namedtuple
from enum import Enum
import numpy as np
from tqdm import tqdm

from ._libtoasty import subsample, mid
from .image import Image
from .pyramid import Pos, tiles_at_depth

HALFPI = 0.5 * np.pi
THREEHALFPI = 1.5 * np.pi
TWOPI = 2 * np.pi


def _arclength(lat1, lon1, lat2, lon2):
    """Compute the length of an arc along the great circle defined by spherical
    latitude and longitude coordinates. Inputs and return value are all in
    radians.

    """
    c = np.sin(lat1) * np.sin(lat2) + np.cos(lon1 - lon2) * np.cos(lat1) * np.cos(lat2)
    return np.arccos(c)


def _spherical_triangle_area(lat1, lon1, lat2, lon2, lat3, lon3):
    """Compute the area of the specified spherical triangle in steradians. Inputs
    are in radians. From https://math.stackexchange.com/a/66731 . My initial
    implementation used unit vectors on the sphere instead of latitudes and
    longitudes; there might be a faster way to do things in lat/lon land.

    """
    c = _arclength(lat1, lon1, lat2, lon2)
    a = _arclength(lat2, lon2, lat3, lon3)
    b = _arclength(lat3, lon3, lat1, lon1)
    s = 0.5 * (a + b + c)
    tane4 = np.sqrt(np.tan(0.5 * s) * np.tan(0.5 * (s - a)) * np.tan(0.5 * (s - b)) * np.tan(0.5 * (s - c)))
    e = 4 * np.arctan(tane4)
    return e


class ToastCoordinateSystem(Enum):
    """
    Different TOAST coordinate systems that are in use.
    """

    ASTRONOMICAL = 'astronomical'
    """The default TOAST coordinate system, where the ``lat = lon = 0`` point
    lies at the middle right edge of the TOAST projection square."""

    PLANETARY = 'planetary'
    """The planetary TOAST coordinate system. This is rotated 180 degrees in
    longitude from the astronomical system, such that the ``lat = lon = 0``
    point lies at the middle left edge of the TOAST projection square."""


Tile = namedtuple('Tile', 'pos corners increasing')

_level1_astronomical_lonlats = np.radians([
    [(0, -90), (90, 0), (0, 90), (180, 0)],
    [(90, 0), (0, -90), (0, 0), (0, 90)],
    [(180, 0), (0, 90), (270, 0), (0, -90)],
    [(0, 90), (0, 0), (0, -90), (270, 0)],
])
_level1_astronomical_lonlats.flags.writeable = False


def _create_level1_tiles(coordsys):
    lonlats = _level1_astronomical_lonlats

    if coordsys == ToastCoordinateSystem.PLANETARY:
        lonlats = lonlats.copy()
        lonlats[...,0] = (lonlats[...,0] + np.pi) % TWOPI

    return [
        Tile(Pos(n=1, x=0, y=0), lonlats[0], True),
        Tile(Pos(n=1, x=1, y=0), lonlats[1], False),
        Tile(Pos(n=1, x=0, y=1), lonlats[2], False),
        Tile(Pos(n=1, x=1, y=1), lonlats[3], True),
    ]


def toast_tile_area(tile):
    """Calculate the area of a TOAST tile in steradians.

    Parameters
    ----------
    tile : :class:`Tile`
        A TOAST tile.

    Returns
    -------
    The area of the tile in steradians.

    Notes
    -----
    This computation is not very fast.

    """
    ul, ur, lr, ll = tile.corners

    if tile.increasing:
        a1 = _spherical_triangle_area(ul[1], ul[0], ur[1], ur[0], ll[1], ll[0])
        a2 = _spherical_triangle_area(ur[1], ur[0], lr[1], lr[0], ll[1], ll[0])
    else:
        a1 = _spherical_triangle_area(ul[1], ul[0], ur[1], ur[0], lr[1], lr[0])
        a2 = _spherical_triangle_area(ul[1], ul[0], ll[1], ll[0], lr[1], lr[0])

    return a1 + a2


def _equ_to_xyz(lat, lon):
    """
    Convert equatorial to cartesian coordinates. Lat and lon are in radians.
    Output is on the unit sphere.

    """
    clat = np.cos(lat)
    return np.array([
        np.cos(lon) * clat,
        np.sin(lat),
        np.sin(lon) * clat,
    ])


def _left_of_half_space_score(point_a, point_b, test_point):
    """
    A variant of WWT Window's IsLeftOfHalfSpace.

    When determining which tile a given RA/Dec lives in, it is inevitable that
    rounding errors can make seem that certain coordinates are not contained by
    *any* tile. Unlike IsLeftOfHalf space, which returns a boolean based on the
    dot product calculated here, we return a number <= 0, where 0 indicates that
    the test point is *definitely* in the left half-space defined by the A and B
    points. Negative values tell us how far into the right space the point is;
    when rounding errors are biting us, that value might be something like
    -1e-16.

    """
    return min(np.dot(np.cross(point_a, point_b), test_point), 0)


def _toast_tile_containment_score(tile, lat, lon):
    """
    Assess whether a TOAST tile contains a given point.

    Parameters
    ----------
    tile : :class:`Tile`
        A TOAST tile
    lat : number
        The latitude (declination) of the point, in radians.
    lon : number
        The longitude (RA) of the point, in radians. This value must
        have already been normalied to lie within the range [0, 2pi]
        (inclusive on both ends.)

    Returns
    -------
    A floating-point "containment score" number. If this number is zero, the
    point definitely lies within the tile. Otherwise, the number will be
    negative, with more negative values indicating a greater distance from the
    point to the nearest tile boundary. Due to inevitable roundoff errors, there
    are situations where, given a certain point and tile, the point "should" be
    contained in the tile, but due to roundoff errors, its score will not be
    exactly zero.

    """
    # Derived from ToastTile.IsPointInTile.

    if tile.pos.n == 0:
        return 0

    # Note that our labeling scheme is different than that used in WWT proper.
    if tile.pos.n == 1:
        if lon >= 0 and lon <= HALFPI and tile.pos.x == 1 and tile.pos.y == 0:
            return 0
        if lon > HALFPI and lon <= np.pi and tile.pos.x == 0 and tile.pos.y == 0:
            return 0
        if lon > np.pi and lon < THREEHALFPI and tile.pos.x == 0 and tile.pos.y == 1:
            return 0
        if lon >= THREEHALFPI and lon <= TWOPI and tile.pos.x == 1 and tile.pos.y == 1:
            return 0
        return -100

    test_point = _equ_to_xyz(lat, lon)
    ul = _equ_to_xyz(tile.corners[0][1], tile.corners[0][0])
    ur = _equ_to_xyz(tile.corners[1][1], tile.corners[1][0])
    lr = _equ_to_xyz(tile.corners[2][1], tile.corners[2][0])
    ll = _equ_to_xyz(tile.corners[3][1], tile.corners[3][0])

    upper = _left_of_half_space_score(ul, ur, test_point)
    right = _left_of_half_space_score(ur, lr, test_point)
    lower = _left_of_half_space_score(lr, ll, test_point)
    left = _left_of_half_space_score(ll, ul, test_point)
    return upper + right + lower + left


def toast_tile_for_point(depth, lat, lon, coordsys=ToastCoordinateSystem.ASTRONOMICAL):
    """
    Identify the TOAST tile at a given depth that contains the given point.

    Parameters
    ----------
    depth : non-negative integer
        The TOAST tile pyramid depth to drill down to. For any given depth,
        there exists a tile containing the input point. As the depth gets
        larger, the precision of the location gets more precise.
    lat : number
        The latitude (declination) of the point, in radians.
    lon : number
        The longitude (RA) of the point, in radians. This value must
        have already been normalized to lie within the range [0, 2pi]
        (inclusive on both ends.)
    coordsys : optional :class:`ToastCoordinateSystem`
        The TOAST coordinate system to use. Default is
        :attr:`ToastCoordinateSystem.ASTRONOMICAL`.

    Returns
    -------
    The :class:`Tile` at the given depth that best contains the specified
    point.

    """
    lon = lon % TWOPI

    if depth == 0:
        return Tile(Pos(n=0, x=0, y=0), (None, None, None, None), False)

    for tile in _create_level1_tiles(coordsys):
        if _toast_tile_containment_score(tile, lat, lon) == 0.:
            break

    while tile.pos.n < depth:
        # Due to inevitable roundoff errors in the tile construction process, it
        # can arise that we find that the point is contained in a certain tile
        # but not contained in any of its children. We deal with this reality by
        # using the "containment score" rather than a binary in/out
        # classification. If no sub-tile has a containment score of zero, we
        # choose whichever tile has the least negative score. In typical
        # roundoff situations that score will be something like -1e-16.
        best_score = -np.inf

        for child in _div4(tile):
            score = _toast_tile_containment_score(child, lat, lon)

            if score == 0.:
                tile = child
                break

            if score > best_score:
                tile = child
                best_score = score

    return tile


def toast_tile_get_coords(tile):
    """
    Get the coordinates of the pixel centers of a TOAST Tile.

    Parameters
    ----------
    tile : :class:`Tile`
        A TOAST tile

    Returns
    -------
    A tuple ``(lons, lats)``, each of which is a 256x256 array of longitudes and
    latitudes of the tile pixel centers, in radians.
    """

    return subsample(
        tile.corners[0],
        tile.corners[1],
        tile.corners[2],
        tile.corners[3],
        256,
        tile.increasing,
    )


def toast_pixel_for_point(depth, lat, lon, coordsys=ToastCoordinateSystem.ASTRONOMICAL):
    """
    Identify the pixel within a TOAST tile at a given depth that contains the
    given point.

    Parameters
    ----------
    depth : non-negative integer
        The TOAST tile pyramid depth to drill down to. For any given depth,
        there exists a tile containing the input point. As the depth gets
        larger, the precision of the location gets more precise.
    lat : number
        The latitude (declination) of the point, in radians.
    lon : number
        The longitude (RA) of the point, in radians. This value must
        have already been normalized to lie within the range [0, 2pi]
        (inclusive on both ends.)
    coordsys : optional :class:`ToastCoordinateSystem`
        The TOAST coordinate system to use. Default is
        :attr:`ToastCoordinateSystem.ASTRONOMICAL`.

    Returns
    -------
    A tuple ``(tile, x, y)``. The *tile* is the :class:`Tile` at the given depth
    that best contains the specified point. The *x* and *y* values are
    floating-point numbers giving the pixel location within the 256×256 tile.
    The returned values are derived from a quadratic fit to the TOAST
    coordinates of the pixels nearest the specified coordinates *lat* and *lon*.

    """
    tile = toast_tile_for_point(depth, lat, lon, coordsys=coordsys)

    # Now that we have the tile, get its pixel locations and identify the pixel
    # that is closest to the input position.

    lons, lats = toast_tile_get_coords(tile)
    dist2 = (lons - lon)**2 + (lats - lat)**2
    min_y, min_x = np.unravel_index(np.argmin(dist2), (256, 256))

    # Now, identify a postage stamp around that best-fit pixel and fit a biquadratic
    # mapping lat/lon to y/x.

    halfsize = 4
    x0 = max(min_x - halfsize, 0)
    y0 = max(min_y - halfsize, 0)
    x1 = min(min_x + halfsize + 1, 256)
    y1 = min(min_y + halfsize + 1, 256)

    dist2_stamp = dist2[y0:y1,x0:x1]
    lons_stamp = lons[y0:y1,x0:x1]
    lats_stamp = lats[y0:y1,x0:x1]

    flat_lons = lons_stamp.flatten()
    flat_lats = lats_stamp.flatten()

    A = np.array([
        flat_lons * 0 + 1,
        flat_lons,
        flat_lats,
        flat_lons**2,
        flat_lons * flat_lats,
        flat_lats**2,
    ]).T

    ygrid, xgrid = np.indices(dist2_stamp.shape)
    x_coeff, _r, _rank, _s = np.linalg.lstsq(A, xgrid.flatten(), rcond=None)
    y_coeff, _r, _rank, _s = np.linalg.lstsq(A, ygrid.flatten(), rcond=None)

    # Evaluate the polynomial to get the refined pixel coordinates.

    pt = np.array([
        1,
        lon,
        lat,
        lon**2,
        lon * lat,
        lat**2,
    ])
    x = np.dot(x_coeff, pt)
    y = np.dot(y_coeff, pt)
    return tile, x0 + x, y0 + y


def _postfix_corner(tile, depth, filter, bottom_only):
    """
    Yield subtiles of a given tile, in postfix (deepest-first) order.

    Parameters
    ----------
    tile : Tile
        Parameters of the current tile.
    depth : int
        The depth to descend to.
    filter : function(Tile)->bool
        A filter function; only tiles for which the function returns True will
        be investigated.
    bottom_only : bool
        If True, only yield tiles at max_depth.

    """
    n = tile[0].n
    if n > depth:
        return

    if n > 1 and not filter(tile):
        return

    for child in _div4(tile):
        for item in _postfix_corner(child, depth, filter, bottom_only):
            yield item

    if n == depth or not bottom_only:
        yield tile


def _div4(tile):
    """Return the four child tiles of an input tile."""
    n, x, y = tile.pos.n, tile.pos.x, tile.pos.y
    ul, ur, lr, ll = tile.corners
    increasing = tile.increasing

    to = mid(ul, ur)
    ri = mid(ur, lr)
    bo = mid(lr, ll)
    le = mid(ll, ul)
    ce = mid(ll, ur) if increasing else mid(ul, lr)

    n += 1
    x *= 2
    y *= 2

    return [
        Tile(Pos(n=n, x=x,     y=y    ), (ul, to, ce, le), increasing),
        Tile(Pos(n=n, x=x + 1, y=y    ), (to, ur, ri, ce), increasing),
        Tile(Pos(n=n, x=x,     y=y + 1), (le, ce, bo, ll), increasing),
        Tile(Pos(n=n, x=x + 1, y=y + 1), (ce, ri, lr, bo), increasing),
    ]


def create_single_tile(pos, coordsys=ToastCoordinateSystem.ASTRONOMICAL):
    """
    Create a single TOAST tile.

    Parameters
    ----------
    pos : :class:`~toasty.pyramid.Pos`
        The position of the tile that will be created. The depth of the
        tile must be at least 1.
    coordsys : optional :class:`ToastCoordinateSystem`
        The TOAST coordinate system to use. Default is
        :attr:`ToastCoordinateSystem.ASTRONOMICAL`.

    Returns
    -------
    :class:`Tile`

    Notes
    -----
    This function should only be used for one-off investigations and debugging.
    It is much more efficient to use :func:`generate_tiles` for bulk computations.
    """

    if pos.n == 0:
        raise ValueError('cannot create a Tile for the n=0 tile')

    children = _create_level1_tiles(coordsys)
    cur_n = 0

    while True:
        cur_n += 1
        ix = (pos.x >> (pos.n - cur_n)) & 0x1
        iy = (pos.y >> (pos.n - cur_n)) & 0x1
        tile = children[iy * 2 + ix]

        if cur_n == pos.n:
            return tile

        children = _div4(tile)


def generate_tiles(depth, bottom_only=True, coordsys=ToastCoordinateSystem.ASTRONOMICAL):
    """Generate a pyramid of TOAST tiles in deepest-first order.

    Parameters
    ----------
    depth : int
        The tile depth to recurse to.
    bottom_only : bool
        If True, then only the lowest tiles will be yielded.
    coordsys : optional :class:`ToastCoordinateSystem`
        The TOAST coordinate system to use. Default is
        :attr:`ToastCoordinateSystem.ASTRONOMICAL`.

    Yields
    ------
    tile : Tile
        An individual tile to process. Tiles are yielded deepest-first.

    The ``n = 0`` depth is not included.

    """
    return generate_tiles_filtered(depth, lambda t: True, bottom_only, coordsys=coordsys)


def generate_tiles_filtered(depth, filter, bottom_only=True, coordsys=ToastCoordinateSystem.ASTRONOMICAL):
    """Generate a pyramid of TOAST tiles in deepest-first order, filtering out subtrees.

    Parameters
    ----------
    depth : int
        The tile depth to recurse to.
    filter : function(Tile)->bool
        A filter function; only tiles for which the function returns True will
        be investigated.
    bottom_only : optional bool
        If True, then only the lowest tiles will be yielded.
    coordsys : optional :class:`ToastCoordinateSystem`
        The TOAST coordinate system to use. Default is
        :attr:`ToastCoordinateSystem.ASTRONOMICAL`.

    Yields
    ------
    tile : Tile
        An individual tile to process. Tiles are yielded deepest-first.

    The ``n = 0`` depth is not included.

    """
    for t in _create_level1_tiles(coordsys):
        for item in _postfix_corner(t, depth, filter, bottom_only):
            yield item


def count_tiles_matching_filter(depth, filter, bottom_only=True, coordsys=ToastCoordinateSystem.ASTRONOMICAL):
    """
    Count the number of tiles matching a filter.

    Parameters
    ----------
    depth : int
        The tile depth to recurse to.
    filter : function(Tile)->bool
        A filter function; only tiles for which the function returns True will
        be investigated.
    bottom_only : bool
        If True, then only the lowest tiles will be processed.
    coordsys : optional :class:`ToastCoordinateSystem`
        The TOAST coordinate system to use. Default is
        :attr:`ToastCoordinateSystem.ASTRONOMICAL`.

    Returns
    ------
    The number of tiles matching the filter. Even if ``bottom_only`` is false,
    the ``n = 0`` tile is not counted.

    Notes
    -----
    This function's call signature and tree-exploration semantics match
    :func:`generate_tiles_filtered`.
    """
    # With a generic filter function, brute force is our only option:
    n = 0

    for _tile in generate_tiles_filtered(depth, filter, bottom_only=bottom_only, coordsys=coordsys):
        n += 1

    return n


def sample_layer(
    pio,
    sampler,
    depth,
    coordsys=ToastCoordinateSystem.ASTRONOMICAL,
    format=None,
    parallel=None,
    cli_progress=False,
):
    """Generate a layer of the TOAST tile pyramid through direct sampling.

    Parameters
    ----------
    pio : :class:`toasty.pyramid.PyramidIO`
        A :class:`~toasty.pyramid.PyramidIO` instance to manage the I/O with
        the tiles in the tile pyramid.
    sampler : callable
        The sampler callable that will produce data for tiling.
    depth : int
        The depth of the layer of the TOAST tile pyramid to generate. The
        number of tiles in each layer is ``4**depth``. Each tile is 256×256
        TOAST pixels, so the resolution of the pixelization at which the
        data will be sampled is a refinement level of ``2**(depth + 8)``.
    coordsys : optional :class:`ToastCoordinateSystem`
        The TOAST coordinate system to use. Default is
        :attr:`ToastCoordinateSystem.ASTRONOMICAL`.
    format : optional :class:`str`
        If provided, override the default data storage format of *pio* with the
        named format, one of the values in ``toasty.image.SUPPORTED_FORMATS``.
    parallel : integer or None (the default)
        The level of parallelization to use. If unspecified, defaults to using
        all CPUs. If the OS does not support fork-based multiprocessing,
        parallel processing is not possible and serial processing will be
        forced. Pass ``1`` to force serial processing.
    cli_progress : optional boolean, defaults False
        If true, a progress bar will be printed to the terminal using tqdm.
    """

    from .par_util import resolve_parallelism
    parallel = resolve_parallelism(parallel)

    if parallel > 1:
        _sample_layer_parallel(pio, format, sampler, depth, coordsys, cli_progress, parallel)
    else:
        _sample_layer_serial(pio, format, sampler, depth, coordsys, cli_progress)


def _sample_layer_serial(pio, format, sampler, depth, coordsys, cli_progress):
    with tqdm(total=tiles_at_depth(depth), disable=not cli_progress) as progress:
        for tile in generate_tiles(depth, bottom_only=True, coordsys=coordsys):
            lon, lat = toast_tile_get_coords(tile)
            sampled_data = sampler(lon, lat)
            pio.write_image(tile.pos, Image.from_array(sampled_data), format=format)
            progress.update(1)

    if cli_progress:
        print()


def _sample_layer_parallel(pio, format, sampler, depth, coordsys, cli_progress, parallel):
    import multiprocessing as mp

    done_event = mp.Event()
    queue = mp.Queue(maxsize = 2 * parallel)
    workers = []

    for _ in range(parallel):
        w = mp.Process(target=_mp_sample_worker, args=(queue, done_event, pio, sampler, format))
        w.daemon = True
        w.start()
        workers.append(w)

    # Send out tiles:

    with tqdm(total=tiles_at_depth(depth), disable=not cli_progress) as progress:
        for tile in generate_tiles(depth, bottom_only=True, coordsys=coordsys):
            queue.put(tile)
            progress.update(1)

    # OK, we're done!

    queue.close()
    queue.join_thread()
    done_event.set()

    for w in workers:
        w.join()

    if cli_progress:
        print()


def _mp_sample_worker(queue, done_event, pio, sampler, format):
    """
    Process tiles on the queue.
    """
    from queue import Empty

    while True:
        try:
            tile = queue.get(True, timeout=1)
        except Empty:
            if done_event.is_set():
                break
            continue

        lon, lat = toast_tile_get_coords(tile)
        sampled_data = sampler(lon, lat)
        pio.write_image(tile.pos, Image.from_array(sampled_data), format=format)


def sample_layer_filtered(
    pio,
    tile_filter,
    sampler,
    depth,
    coordsys=ToastCoordinateSystem.ASTRONOMICAL,
    parallel=None,
    cli_progress=False,
):
    """Populate a subset of a layer of the TOAST tile pyramid through direct sampling.

    Parameters
    ----------
    pio : :class:`toasty.pyramid.PyramidIO`
        A :class:`~toasty.pyramid.PyramidIO` instance to manage the I/O with
        the tiles in the tile pyramid.
    tile_filter : callable
        A tile filtering function, suitable for passing to
        :func:`toasty.toast.generate_tiles_filtered`.
    sampler : callable
        The sampler callable that will produce data for tiling.
    depth : int
        The depth of the layer of the TOAST tile pyramid to generate. The
        number of tiles in each layer is ``4**depth``. Each tile is 256×256
        TOAST pixels, so the resolution of the pixelization at which the
        data will be sampled is a refinement level of ``2**(depth + 8)``.
    coordsys : optional :class:`ToastCoordinateSystem`
        The TOAST coordinate system to use. Default is
        :attr:`ToastCoordinateSystem.ASTRONOMICAL`.
    parallel : integer or None (the default)
        The level of parallelization to use. If unspecified, defaults to using
        all CPUs. If the OS does not support fork-based multiprocessing,
        parallel processing is not possible and serial processing will be
        forced. Pass ``1`` to force serial processing.
    cli_progress : optional boolean, defaults False
        If true, a progress bar will be printed to the terminal using tqdm.
    """

    from .par_util import resolve_parallelism
    parallel = resolve_parallelism(parallel)

    if parallel > 1:
        _sample_filtered_parallel(pio, tile_filter, sampler, depth, coordsys, cli_progress, parallel)
    else:
        _sample_filtered_serial(pio, tile_filter, sampler, depth, coordsys, cli_progress)


def _sample_filtered_serial(pio, tile_filter, sampler, depth, coordsys, cli_progress):
    n_todo = count_tiles_matching_filter(depth, tile_filter, bottom_only=True, coordsys=coordsys)

    with tqdm(total=n_todo, disable=not cli_progress) as progress:
        for tile in generate_tiles_filtered(depth, tile_filter, bottom_only=True, coordsys=coordsys):
            lon, lat = toast_tile_get_coords(tile)
            sampled_data = sampler(lon, lat)
            img = Image.from_array(sampled_data)

            with pio.update_image(tile.pos, masked_mode=img.mode, default='masked') as basis:
                img.update_into_maskable_buffer(basis, slice(None), slice(None), slice(None), slice(None))

            progress.update(1)

    if cli_progress:
        print()

    # do not clean lockfiles, for HPC contexts where we're processing different
    # chunks in parallel.


def _sample_filtered_parallel(pio, tile_filter, sampler, depth, coordsys, cli_progress, parallel):
    import multiprocessing as mp

    n_todo = count_tiles_matching_filter(depth, tile_filter, bottom_only=True, coordsys=coordsys)

    done_event = mp.Event()
    queue = mp.Queue(maxsize = 2 * parallel)
    workers = []

    for _ in range(parallel):
        w = mp.Process(target=_mp_sample_filtered, args=(queue, done_event, pio, sampler))
        w.daemon = True
        w.start()
        workers.append(w)

    # Here we go:

    with tqdm(total=n_todo, disable=not cli_progress) as progress:
        for tile in generate_tiles_filtered(depth, tile_filter, bottom_only=True, coordsys=coordsys):
            queue.put(tile)
            progress.update(1)

    # OK, we're done!

    queue.close()
    queue.join_thread()
    done_event.set()

    for w in workers:
        w.join()

    if cli_progress:
        print()

    # do not clean lockfiles, for HPC contexts where we're processing different
    # chunks in parallel.


def _mp_sample_filtered(queue, done_event, pio, sampler):
    """
    Process tiles on the queue.

    We use ``pio.update_image`` in case we are in an HPC-type context where
    other chunks may be trying to write out the populate the same tiles as us at
    the same time.
    """

    from queue import Empty

    while True:
        try:
            tile = queue.get(True, timeout=1)
        except Empty:
            if done_event.is_set():
                break
            continue

        lon, lat = toast_tile_get_coords(tile)
        sampled_data = sampler(lon, lat)
        img = Image.from_array(sampled_data)

        with pio.update_image(tile.pos, masked_mode=img.mode, default='masked') as basis:
            img.update_into_maskable_buffer(basis, slice(None), slice(None), slice(None), slice(None))



# -*- mode: python; coding: utf-8 -*-
# Copyright 2021 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Collections of related input images.

Some Toasty processing tasks operate over collections of input images that are
related in some way. This module provides standardized mechanisms for
manipulating such collections. In particular, it provides a framework for
scanning over image "descriptions" without reading in the complete image data.
This can be useful because many collection-related operations want to do an
initial pass over the collection to gather some global information, then a
second pass with the actual data processing.
"""

__all__ = '''
ImageCollection
RubinDirectoryCollection
SimpleFitsCollection
'''.split()

from abc import ABC
from glob import glob
import numpy as np
from os.path import join
import warnings

from .image import Image, ImageDescription, ImageMode
from .study import StudyTiling


class ImageCollection(ABC):
    def descriptions(self):
        """
        Generate a sequence of :class:`toasty.image.ImageDescription` items
        associated with this collection.

        Each description will have an added string attribute ``collection_id``
        that gives a unique textual identifer for the item in the collection.

        Unlike :meth:`ImageCollection.images`, this function does cause the full
        data for each image to be loaded.
        """
        raise NotImplementedError()

    def images(self):
        """
        Generate a sequence of :class:`toasty.image.Image` items associated with
        this collection.

        Each image will have an added string attribute ``collection_id`` that
        gives a unique textual identifer for the item in the collection.
        """
        raise NotImplementedError()


class SimpleFitsCollection(ImageCollection):
    def __init__(self, paths, hdu_index=None, blankval=None):
        self._paths = list(paths)
        self._hdu_index = hdu_index
        self._blankval = blankval

    def _load(self, actually_load_data):
        from astropy.io import fits
        from astropy.wcs import WCS

        for fits_path in self._paths:
            with fits.open(fits_path) as hdul:
                if self._hdu_index is not None:
                    hdu = hdul[self._hdu_index]
                else:
                    for hdu in hdul:
                        if (hasattr(hdu, 'shape') and len(hdu.shape) > 1
                                and type(hdu) is not fits.hdu.table.BinTableHDU):
                            break

                if type(hdu) is fits.hdu.table.BinTableHDU:
                    raise Exception(f'cannot process input `{fits_path}`: Did not find any HDU with image data')
                wcs = WCS(hdu.header)
                shape = hdu.shape

                # We need to make sure the data are 2D celestial, since that's
                # what our image code and `reproject` (if it's being used) expect.

                full_wcs = None
                keep_axes = None

                if wcs.naxis != 2:
                    if not wcs.has_celestial:
                        raise Exception(f'cannot process input `{fits_path}`: WCS cannot be reduced to 2D celestial')

                    full_wcs = wcs
                    wcs = full_wcs.celestial

                    # note: get_axis_types returns axes in FITS order, innermost first
                    keep_axes = [t.get('coordinate_type') == 'celestial' for t in full_wcs.get_axis_types()[::-1]]

                    for keep, axlen in zip(keep_axes, shape):
                        if not keep and axlen != 1:
                            raise Exception(f'cannot process input `{fits_path}`: found a non-celestial axis with size other than 1')

                # OK, almost there. Are we loading data or just the descriptor?

                if actually_load_data:
                    data = hdu.data

                    if full_wcs is not None:  # need to subset?
                        data = data[tuple(slice(None) if k else 0 for k in keep_axes)]

                    if self._blankval is not None:
                        data[data == self._blankval] = np.nan

                    result = Image.from_array(data, wcs=wcs, default_format='fits')
                else:
                    if full_wcs is not None:  # need to subset?
                        shape = tuple(t[1] for t in zip(keep_axes, shape) if t[0])

                    if hasattr(hdu, 'dtype'):
                        mode = ImageMode.from_array_info(shape, hdu.dtype)
                    else:
                        mode = None  # CompImageHDU doesn't have dtype

                    result = ImageDescription(mode=mode, shape=shape, wcs=wcs)

                result.collection_id = fits_path
                yield result

    def descriptions(self):
        return self._load(False)

    def images(self):
        return self._load(True)


class RubinDirectoryCollection(ImageCollection):
    """
    Load up imagery from a directory containing FITS files capturing a Rubin
    Observatory tract.

    The directory will be searched for files whose names end in ``.fits``. In
    each of those files, the HDUs beyond the first will be treated as separate
    science images that should be individually loaded. The images will be
    trimmed according to their ``DATASEC`` specification before being returned.

    This class requires the ``ccdproc``  package to trim FITS CCD datasets.
    """

    def __init__(self, dirname, unit=None):
        self._dirname = dirname
        self._unit = unit

    def _load(self, actually_load_data):
        from astropy.io import fits
        from astropy.nddata import ccddata
        import ccdproc

        for fits_path in glob(join(self._dirname, '*.fits')):
            # `astropy.nddata.ccddata.fits_ccddata_reader` only opens FITS from
            # filenames, not from an open HDUList, which means that creating
            # multiple CCDDatas from the same FITS file rapidly becomes
            # inefficient. So, we emulate its logic.

            with fits.open(fits_path) as hdu_list:
                for idx, hdu in enumerate(hdu_list):
                    if idx == 0:
                        header0 = hdu.header
                    else:
                        hdr = hdu.header
                        hdr.extend(header0, unique=True)

                        # This ccddata function often generates annoying warnings
                        with warnings.catch_warnings():
                            warnings.simplefilter('ignore')
                            hdr, wcs = ccddata._generate_wcs_and_update_header(hdr)

                        # We can't use `ccdproc.trim_image()` without having a
                        # CCDData in hand, so we have to create a fake empty
                        # array even when we're not actually loading any data.
                        #
                        # Note: we skip all the unit-handling logic from
                        # `fits_ccddata_reader` here basically since the LSST
                        # sim data I'm using don't have anything useful.

                        if actually_load_data:
                            data = hdu.data

                            if data.dtype.kind == 'i':
                                data = data.astype(np.float32)
                        else:
                            data = np.empty(hdu.shape, dtype=np.void)

                        ccd = ccddata.CCDData(data, meta=hdr, unit=self._unit, wcs=wcs)
                        ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['DATASEC'])
                        data = ccd.data
                        shape = data.shape
                        wcs = ccd.wcs

                        if actually_load_data:
                            mode = ImageMode.from_array_info(shape, data.dtype)
                        elif hasattr(hdu, 'dtype'):
                            mode = ImageMode.from_array_info(shape, hdu.dtype)
                        else:
                            mode = None  # CompImageHDU doesn't have dtype

                        if actually_load_data:
                            result = Image.from_array(data, wcs=wcs, default_format='fits')
                        else:
                            result = ImageDescription(mode=mode, shape=shape, wcs=wcs)

                        result.collection_id = f'{fits_path}:{idx}'
                        yield result

    def descriptions(self):
        return self._load(False)

    def images(self):
        return self._load(True)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2021 the AAS WorldWide Telescope project.
# Licensed under the MIT License.

"""Entrypoint for the "toasty" command-line interface.

"""

from __future__ import absolute_import, division, print_function

__all__ = '''
die
entrypoint
warn
'''.split()

import argparse
import os.path
import sys
import warnings

from wwt_data_formats.cli import EnsureGlobsExpandedAction


# General CLI utilities

def die(msg):
    print('error:', msg, file=sys.stderr)
    sys.exit(1)

def warn(msg):
    print('warning:', msg, file=sys.stderr)


# "cascade" subcommand

def cascade_getparser(parser):
    parser.add_argument(
        '--parallelism', '-j',
        metavar = 'COUNT',
        type = int,
        help = 'The parallelization level (default: use all CPUs; specify `1` to force serial processing)',
    )
    parser.add_argument(
        '--format', '-f',
        metavar = 'FORMAT',
        default = None,
        choices = ['png', 'jpg', 'npy', 'fits'],
        help = 'The format of data files to cascade. If not specified, this will be guessed.',
    )
    parser.add_argument(
        '--start',
        metavar = 'DEPTH',
        type = int,
        help = 'The depth of the TOAST layer to start the cascade',
    )
    parser.add_argument(
        'pyramid_dir',
        metavar = 'DIR',
        help = 'The directory containing the tile pyramid to cascade',
    )


def cascade_impl(settings):
    from .image import ImageMode
    from .merge import averaging_merger, cascade_images
    from .pyramid import PyramidIO

    pio = PyramidIO(settings.pyramid_dir, default_format=settings.format)

    start = settings.start
    if start is None:
        die('currently, you must specify the start layer with the --start option')

    cascade_images(
        pio,
        start,
        averaging_merger,
        parallel=settings.parallelism,
        cli_progress=True
    )


# "check-avm" subcommand

def check_avm_getparser(parser):
    parser.add_argument(
        '--print', '-p',
        dest = 'print_avm',
        action = 'store_true',
        help = 'Print the AVM data if present',
    )
    parser.add_argument(
        '--exitcode',
        action = 'store_true',
        help = 'Exit with code 1 if no AVM tags are present',
    )
    parser.add_argument(
        'image',
        metavar = 'PATH',
        help = 'An image to check for AVM tags',
    )


def check_avm_impl(settings):
    try:
        from pyavm import AVM, exceptions

        SYNTAX_EXCEPTIONS = (
            exceptions.AVMListLengthError,
            exceptions.AVMItemNotInControlledVocabularyError,
            exceptions.AVMEmptyValueError,
        )
    except ImportError:
        die('cannot check AVM tags: you must install the `pyavm` package')

    try:
        # PyAVM can issue a lot of warnings that aren't helpful to the user.
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            avm = AVM.from_image(settings.image)
    except exceptions.NoXMPPacketFound:
        print(f'{settings.image}: no AVM tags')
        sys.exit(1 if settings.exitcode else 0)
    except SYNTAX_EXCEPTIONS as e:
        print(f'{settings.image}: AVM data are present but malformed: {e} ({e.__class__.__name__})')
        sys.exit(1)
    except IOError as e:
        die(f'error probing AVM tags of `{settings.image}`: {e}')

    if settings.print_avm:
        print(f'{settings.image}: AVM tags are present')
        print()
        print(avm)
    else:
        print(f'{settings.image}: AVM tags are present (use `--print` to display them)')


# "make_thumbnail" subcommand

def make_thumbnail_getparser(parser):
    from .image import ImageLoader
    ImageLoader.add_arguments(parser)

    parser.add_argument(
        'imgpath',
        metavar = 'IN-PATH',
        help = 'The image file to be thumbnailed',
    )
    parser.add_argument(
        'outpath',
        metavar = 'OUT-PATH',
        help = 'The location of the new thumbnail file',
    )


def make_thumbnail_impl(settings):
    from .image import ImageLoader

    olp = settings.outpath.lower()
    if not (olp.endswith('.jpg') or olp.endswith('.jpeg')):
        warn('saving output in JPEG format even though filename is "{}"'.format(settings.outpath))

    img = ImageLoader.create_from_args(settings).load_path(settings.imgpath)
    thumb = img.make_thumbnail_bitmap()

    with open(settings.outpath, 'wb') as f:
        thumb.save(f, format='JPEG')


# "pipeline" subcommands

from .pipeline.cli import pipeline_getparser, pipeline_impl


# "tile_allsky" subcommand

def tile_allsky_getparser(parser):
    from .image import ImageLoader
    ImageLoader.add_arguments(parser)

    parser.add_argument(
        '--name',
        metavar = 'NAME',
        default = 'Toasty',
        help = 'The image name to embed in the output WTML file (default: %(default)s)',
    )
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        default = '.',
        help = 'The root directory of the output tile pyramid (default: %(default)s)',
    )
    parser.add_argument(
        '--placeholder-thumbnail',
        action = 'store_true',
        help = 'Do not attempt to thumbnail the input image -- saves memory for large inputs',
    )
    parser.add_argument(
        '--projection',
        metavar = 'PROJTYPE',
        default = 'plate-carree',
        help = 'The projection type of the input image (default: %(default)s; choices: %(choices)s)',
        choices = ['plate-carree', 'plate-carree-galactic', 'plate-carree-ecliptic', 'plate-carree-planet', 'plate-carree-panorama'],
    )
    parser.add_argument(
        '--parallelism', '-j',
        metavar = 'COUNT',
        type = int,
        help = 'The parallelization level (default: use all CPUs if OS supports; specify `1` to force serial processing)',
    )
    parser.add_argument(
        'imgpath',
        metavar = 'PATH',
        help = 'The image file to be tiled',
    )
    parser.add_argument(
        'depth',
        metavar = 'DEPTH',
        type = int,
        help = 'The depth of the TOAST layer to sample',
    )


def tile_allsky_impl(settings):
    from .builder import Builder
    from .image import ImageLoader
    from .pyramid import PyramidIO

    img = ImageLoader.create_from_args(settings).load_path(settings.imgpath)
    pio = PyramidIO(settings.outdir)
    is_planet = False
    is_pano = False

    if settings.projection == 'plate-carree':
        from .samplers import plate_carree_sampler
        sampler = plate_carree_sampler(img.asarray())
    elif settings.projection == 'plate-carree-galactic':
        from .samplers import plate_carree_galactic_sampler
        sampler = plate_carree_galactic_sampler(img.asarray())
    elif settings.projection == 'plate-carree-ecliptic':
        from .samplers import plate_carree_ecliptic_sampler
        sampler = plate_carree_ecliptic_sampler(img.asarray())
    elif settings.projection == 'plate-carree-planet':
        from .samplers import plate_carree_planet_sampler
        sampler = plate_carree_planet_sampler(img.asarray())
        is_planet = True
    elif settings.projection == 'plate-carree-panorama':
        from .samplers import plate_carree_sampler
        sampler = plate_carree_sampler(img.asarray())
        is_pano = True
    else:
        die('the image projection type {!r} is not recognized'.format(settings.projection))

    builder = Builder(pio)

    # Do the thumbnail first since for large inputs it can be the memory high-water mark!
    if settings.placeholder_thumbnail:
        builder.make_placeholder_thumbnail()
    else:
        builder.make_thumbnail_from_other(img)

    builder.toast_base(
        sampler,
        settings.depth,
        is_planet=is_planet,
        is_pano=is_pano,
        parallel=settings.parallelism,
        cli_progress=True,
    )
    builder.set_name(settings.name)
    builder.write_index_rel_wtml()

    print(f'Successfully tiled input "{settings.imgpath}" at level {builder.imgset.tile_levels}.')
    print('To create parent tiles, consider running:')
    print()
    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')


# "tile_healpix" subcommand

def tile_healpix_getparser(parser):
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        default = '.',
        help = 'The root directory of the output tile pyramid',
    )
    parser.add_argument(
        'fitspath',
        metavar = 'PATH',
        help = 'The HEALPix FITS file to be tiled',
    )
    parser.add_argument(
        'depth',
        metavar = 'DEPTH',
        type = int,
        help = 'The depth of the TOAST layer to sample',
    )


def tile_healpix_impl(settings):
    from .builder import Builder
    from .image import ImageMode
    from .pyramid import PyramidIO
    from .samplers import healpix_fits_file_sampler

    pio = PyramidIO(settings.outdir, default_format='npy')
    sampler = healpix_fits_file_sampler(settings.fitspath)
    builder = Builder(pio)
    builder.toast_base(sampler, settings.depth)
    builder.write_index_rel_wtml()

    print(f'Successfully tiled input "{settings.fitspath}" at level {builder.imgset.tile_levels}.')
    print('To create parent tiles, consider running:')
    print()
    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')


# "tile_multi_tan" subcommand

def tile_multi_tan_getparser(parser):
    parser.add_argument(
        '--parallelism', '-j',
        metavar = 'COUNT',
        type = int,
        help = 'The parallelization level (default: use all CPUs; specify `1` to force serial processing)',
    )
    parser.add_argument(
        '--hdu-index',
        metavar = 'INDEX',
        type = int,
        default = 0,
        help = 'Which HDU to load in each input FITS file',
    )
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        default = '.',
        help = 'The root directory of the output tile pyramid',
    )
    parser.add_argument(
        'paths',
        metavar = 'PATHS',
        action = EnsureGlobsExpandedAction,
        nargs = '+',
        help = 'The FITS files with image data',
    )

def tile_multi_tan_impl(settings):
    from .builder import Builder
    from .collection import SimpleFitsCollection
    from .image import ImageMode
    from .multi_tan import MultiTanProcessor
    from .pyramid import PyramidIO

    pio = PyramidIO(settings.outdir, default_format='fits')
    builder = Builder(pio)

    collection = SimpleFitsCollection(settings.paths, hdu_index=settings.hdu_index)

    mtp = MultiTanProcessor(collection)
    mtp.compute_global_pixelization(builder)
    mtp.tile(
        pio,
        parallel=settings.parallelism,
        cli_progress=True,
    )
    builder.write_index_rel_wtml()

    print(f'Successfully tiled inputs at level {builder.imgset.tile_levels}.')
    print('To create parent tiles, consider running:')
    print()
    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')


# "tile_study" subcommand

def tile_study_getparser(parser):
    from .image import ImageLoader
    ImageLoader.add_arguments(parser)

    parser.add_argument(
        '--name',
        metavar = 'NAME',
        default = 'Toasty',
        help = 'The image name to embed in the output WTML file (default: %(default)s)',
    )
    parser.add_argument(
        '--avm',
        action = 'store_true',
        help = 'Expect the input image to have AVM positioning tags',
    )
    parser.add_argument(
        '--placeholder-thumbnail',
        action = 'store_true',
        help = 'Do not attempt to thumbnail the input image -- saves memory for large inputs',
    )
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        default = '.',
        help = 'The root directory of the output tile pyramid',
    )
    parser.add_argument(
        'imgpath',
        metavar = 'PATH',
        help = 'The study image file to be tiled',
    )


def tile_study_impl(settings):
    from .builder import Builder
    from .image import ImageLoader
    from .pyramid import PyramidIO

    img = ImageLoader.create_from_args(settings).load_path(settings.imgpath)
    pio = PyramidIO(settings.outdir, default_format=img.default_format)
    builder = Builder(pio)

    # First, prepare the tiling, because we can't correctly apply WCS until we
    # know the details of the tiling parameters

    tiling = builder.prepare_study_tiling(img)

    # Now deal with the WCS

    if settings.avm:
        # We don't *always* check for AVM because pyavm prints out a lot of junk
        # and may not be installed.
        try:
            from pyavm import AVM
        except ImportError:
            die('cannot use AVM data: you must install the `pyavm` package')

        try:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                avm = AVM.from_image(settings.imgpath)
        except Exception:
            print(f'error: failed to read AVM tags of input `{settings.imgpath}`', file=sys.stderr)
            raise

        builder.apply_avm_info(avm, img.width, img.height)
    elif img.wcs is not None:
        # Study images must have negative parity.
        img.ensure_negative_parity()
        builder.apply_wcs_info(img.wcs, img.width, img.height)
    else:
        builder.default_tiled_study_astrometry()

    # Do the thumbnail before the main image since for large inputs it can be
    # the memory high-water mark!
    if settings.placeholder_thumbnail:
        builder.make_placeholder_thumbnail()
    else:
        builder.make_thumbnail_from_other(img)

    # Finally, actually tile the image
    builder.execute_study_tiling(img, tiling, cli_progress=True)

    # Final metadata
    builder.set_name(settings.name)
    builder.write_index_rel_wtml()

    print(f'Successfully tiled input "{settings.imgpath}" at level {builder.imgset.tile_levels}.')
    print('To create parent tiles, consider running:')
    print()
    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')


# "tile_wwtl" subcommand

def tile_wwtl_getparser(parser):
    from .image import ImageLoader
    ImageLoader.add_arguments(parser)

    parser.add_argument(
        '--name',
        metavar = 'NAME',
        default = 'Toasty',
        help = 'The image name to embed in the output WTML file (default: %(default)s)',
    )
    parser.add_argument(
        '--placeholder-thumbnail',
        action = 'store_true',
        help = 'Do not attempt to thumbnail the input image -- saves memory for large inputs',
    )
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        default = '.',
        help = 'The root directory of the output tile pyramid',
    )
    parser.add_argument(
        'wwtl_path',
        metavar = 'WWTL-PATH',
        help = 'The WWTL layer file to be processed',
    )


def tile_wwtl_impl(settings):
    from .builder import Builder
    from .image import ImageLoader
    from .pyramid import PyramidIO

    pio = PyramidIO(settings.outdir)
    builder = Builder(pio)
    img = builder.load_from_wwtl(settings, settings.wwtl_path, cli_progress=True)

    # Do the thumbnail first since for large inputs it can be the memory high-water mark!
    if settings.placeholder_thumbnail:
        builder.make_placeholder_thumbnail()
    else:
        builder.make_thumbnail_from_other(img)

    builder.set_name(settings.name)
    builder.write_index_rel_wtml()

    print(f'Successfully tiled input "{settings.wwtl_path}" at level {builder.imgset.tile_levels}.')
    print('To create parent tiles, consider running:')
    print()
    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')


# "transform" subcommand

def transform_getparser(parser):
    subparsers = parser.add_subparsers(dest='transform_command')

    parser = subparsers.add_parser('fx3-to-rgb')
    parser.add_argument(
        '--parallelism', '-j',
        metavar = 'COUNT',
        type = int,
        help = 'The parallelization level (default: use all CPUs; specify `1` to force serial processing)',
    )
    parser.add_argument(
        '--start',
        metavar = 'DEPTH',
        type = int,
        help = 'How deep in the tile pyramid to proceed wth the transformation',
    )
    parser.add_argument(
        '--clip', '-c',
        metavar = 'NUMBER',
        type = float,
        default = 1.0,
        help = 'The level at which to start flipping the floating-point data',
    )
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        help = 'Output transformed data into this directory, instead of operating in-place',
    )
    parser.add_argument(
        'pyramid_dir',
        metavar = 'DIR',
        help = 'The directory containing the tile pyramid to transform',
    )

    parser = subparsers.add_parser('u8-to-rgb')
    parser.add_argument(
        '--parallelism', '-j',
        metavar = 'COUNT',
        type = int,
        help = 'The parallelization level (default: use all CPUs; specify `1` to force serial processing)',
    )
    parser.add_argument(
        '--start',
        metavar = 'DEPTH',
        type = int,
        help = 'How deep in the tile pyramid to proceed wth the transformation',
    )
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        help = 'Output transformed data into this directory, instead of operating in-place',
    )
    parser.add_argument(
        'pyramid_dir',
        metavar = 'DIR',
        help = 'The directory containing the tile pyramid to transform',
    )


def transform_impl(settings):
    from .pyramid import PyramidIO

    if settings.transform_command is None:
        print('Run the "transform" command with `--help` for help on its subcommands')
        return

    if settings.transform_command == 'fx3-to-rgb':
        pio = PyramidIO(settings.pyramid_dir)
        pio_out = None

        if settings.outdir is not None:
            pio_out = PyramidIO(settings.outdir)

        from .transform import f16x3_to_rgb
        f16x3_to_rgb(
            pio, settings.start,
            clip = settings.clip,
            pio_out = pio_out,
            parallel = settings.parallelism,
            cli_progress = True,
        )
    elif settings.transform_command == 'u8-to-rgb':
        pio = PyramidIO(settings.pyramid_dir)
        pio_out = None

        if settings.outdir is not None:
            pio_out = PyramidIO(settings.outdir)

        from .transform import u8_to_rgb
        u8_to_rgb(
            pio, settings.start,
            pio_out = pio_out,
            parallel = settings.parallelism,
            cli_progress = True,
        )
    else:
        die('unrecognized "transform" subcommand ' + settings.transform_command)


# The CLI driver:

def entrypoint(args=None):
    """The entrypoint for the \"toasty\" command-line interface.

    Parameters
    ----------
    args : iterable of str, or None (the default)
      The arguments on the command line. The first argument should be
      a subcommand name or global option; there is no ``argv[0]``
      parameter.

    """
    # Set up the subcommands from globals()

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="subcommand")
    commands = set()

    for py_name, value in globals().items():
        if py_name.endswith('_getparser'):
            cmd_name = py_name[:-10].replace('_', '-')
            subparser = subparsers.add_parser(cmd_name)
            value(subparser)
            commands.add(cmd_name)

    # What did we get?

    settings = parser.parse_args(args)

    if settings.subcommand is None:
        print('Run me with --help for help. Allowed subcommands are:')
        print()
        for cmd in sorted(commands):
            print('   ', cmd)
        return

    py_name = settings.subcommand.replace('-', '_')

    impl = globals().get(py_name + '_impl')
    if impl is None:
        die('no such subcommand "{}"'.format(settings.subcommand))

    # OK to go!

    impl(settings)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2013-2019 Chris Beaumont and the AAS WorldWide Telescope project
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function



# -*- mode: python; coding: utf-8 -*-
# Copyright 2021 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Special handling for JPEG2000 files.

This is very primitive support. Implemented for:

https://planetarymaps.usgs.gov/mosaic/Lunar_Kaguya_TC_Ortho_Global_4096ppd_v02.jp2.html

JPEG2000's can be tiled. We refer to these tiles as "chunks" to try to avoid
confusion with Toasty's tiling process (although perhaps this just makes things
worse).
"""

from __future__ import absolute_import, division, print_function

__all__ = '''
ChunkedJPEG2000Reader
HAS_JPEG2000
'''.split()

try:
    import glymur
    HAS_JPEG2000 = True
except ImportError:
    HAS_JPEG2000 = False


class ChunkedJPEG2000Reader(object):
    """
    A simple reader for tiled JPEG2000 files.

    Parameters
    ----------
    path : path-like
        The path of the JP2 input file
    fake_data : optional bool
        The default is false. If true, no data are read from the file. Instead
        an array of 255's is returned. This can be useful for testing, where in
        my large test case (the Kaguya JP2) it can take 15 minutes to read a
        single tile from the file.
    """

    def __init__(self, path, fake_data=False):
        if not HAS_JPEG2000:
            raise Exception('you must install the `glymur` Python package to process JPEG2000 files')

        self._jp2 = glymur.Jp2k(path)
        self._tile_shape = None
        self._fake_data = fake_data

        for seg in self._jp2.codestream.segment:
            if seg.marker_id == 'SIZ':
                self._tile_shape = (seg.ytsiz, seg.xtsiz)

    @property
    def shape(self):
        """
        A tuple of either ``(height, width)`` or ``(height, width, n_components)``.
        """
        return self._jp2.shape

    @property
    def n_chunks(self):
        """
        The number of chunks (tiles) inside the JPEG2000 file.
        """
        th, tw = self._tile_shape
        return ((self._jp2.shape[0] + th - 1) // th) * ((self._jp2.shape[1] + tw - 1) // tw)

    def chunk_spec(self, ichunk):
        """
        Get information about an image chunk that can be loaded.

        Parameters
        ----------
        ichunk : :class:`int`
            The index of the chunk to query, between 0 and ``self.n_chunks - 1``
            (inclusive).

        Returns
        -------
        Tuple of ``(x, y, width, height)``, where the coordinates refer to
        the top-left corner of the chunk relative to the overall image
        (0-based).
        """
        if ichunk < 0 or ichunk >= self.n_chunks:
            raise ValueError(f'got ichunk={ichunk!r}; should be between 0 and {self.n_chunks-1}')

        th, tw = self._tile_shape
        gh, gw = self._jp2.shape[:2]
        chunks_per_row = (gw + tw - 1) // tw
        icol = ichunk // chunks_per_row
        irow = ichunk % chunks_per_row

        x0 = tw * irow
        x1 = min(tw * (irow + 1), gw)
        chunk_width = x1 - x0

        y0 = th * icol
        y1 = min(th * (icol + 1), gh)
        chunk_height = y1 - y0

        return x0, y0, chunk_width, chunk_height

    def chunk_data(self, ichunk):
        """
        Load an image chunk.

        Parameters
        ----------
        ichunk : :class:`int`
            The index of the chunk to query, between 0 and ``self.n_chunks - 1``
            (inclusive).

        Returns
        -------
        A numpy array of image data whose overall shape depends on the image
        format (see :attr:`shape`), but whose width and height will agree with
        the chunk specification.

        Notes
        -----
        This can be extremely slow. With the large chunks used by the Kaguya JPEG2000
        file, it can take ~15 minutes to load a typical chunk.
        """
        # The only way that the glymur package makes it possible to read
        # explicitly tile-by-tile is through a deprecated API, but if we use the
        # newer indexing API, it seems that we get the same effect (including
        # performance) with the following approach.
        x0, y0, w, h = self.chunk_spec(ichunk)

        if self._fake_data:
            import numpy as np
            a = np.empty((h, w), dtype=np.uint8)
            a.fill(255)
            return a

        return self._jp2[y0:y0+h,x0:x0+w]



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020-2021 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Within-tile data transformations. Mainly, transforming floating-point data to
RGB.
"""

__all__ = '''
f16x3_to_rgb
u8_to_rgb
'''.split()

from astropy import visualization as viz
import numpy as np
from tqdm import tqdm

from .image import ImageMode, Image
from .pyramid import depth2tiles, generate_pos


# Generalized transform machinery, with both serial and parallel support

def _do_a_transform(pio, depth, make_buf, do_one, pio_out=None, parallel=None, cli_progress=False):
    from .par_util import resolve_parallelism
    parallel = resolve_parallelism(parallel)

    if pio_out is None:
        pio_out = pio

    if parallel > 1:
        _transform_parallel(pio, pio_out, depth, make_buf, do_one, cli_progress, parallel)
    else:
        buf = make_buf()

        with tqdm(total=depth2tiles(depth), disable=not cli_progress) as progress:
            for pos in generate_pos(depth):
                do_one(buf, pos, pio, pio_out)
                progress.update(1)

        if cli_progress:
            print()


def _transform_parallel(pio_in, pio_out, depth, make_buf, do_one, cli_progress, parallel):
    import multiprocessing as mp

    # Start up the workers

    done_event = mp.Event()
    queue = mp.Queue(maxsize = 16 * parallel)
    workers = []

    for _ in range(parallel):
        w = mp.Process(target=_transform_mp_worker, args=(queue, done_event, pio_in, pio_out, make_buf, do_one))
        w.daemon = True
        w.start()
        workers.append(w)

    # Send out them tiles

    with tqdm(total=depth2tiles(depth), disable=not cli_progress) as progress:
        for pos in generate_pos(depth):
              queue.put(pos)
              progress.update(1)

    # All done

    queue.close()
    queue.join_thread()
    done_event.set()

    for w in workers:
        w.join()

    if cli_progress:
        print()


def _transform_mp_worker(queue, done_event, pio_in, pio_out, make_buf, do_one):
    """
    Do the colormapping.
    """
    from queue import Empty

    buf = make_buf()

    while True:
        try:
            pos = queue.get(True, timeout=1)
        except Empty:
            if done_event.is_set():
                break
            continue

        do_one(buf, pos, pio_in, pio_out)


# float-to-RGB(A), with a generalized float-to-unit transform

def f16x3_to_rgb(pio, start_depth, clip=1, **kwargs):
    transform = viz.SqrtStretch() + viz.ManualInterval(0, clip)
    _float_to_rgb(pio, start_depth, transform, **kwargs)


def _float_to_rgb(pio_in, depth, transform, out_format='png', **kwargs):
    make_buf = lambda: np.empty((256, 256, 3), dtype=np.uint8)
    do_one = lambda buf, pos, pio_in, pio_out: _float_to_rgb_do_one(buf, pos, pio_in, pio_out, transform, out_format)
    _do_a_transform(pio_in, depth, make_buf, do_one, **kwargs)


def _float_to_rgb_do_one(buf, pos, pio_in, pio_out, transform, out_format):
    img = pio_in.read_image(pos, format='npy')
    if img is None:
        return

    mapped = transform(img.asarray())

    if mapped.ndim == 2:
        # TODO: allow real colormaps
        valid = np.isfinite(mapped)
        mapped = mapped.reshape((256, 256, 1))
    else:
        valid = np.all(np.isfinite(mapped), axis=2)

    mapped[~valid] = 0
    mapped = np.clip(mapped * 255, 0, 255).astype(np.uint8)
    buf[...] = mapped
    rgb = Image.from_array(buf)
    pio_out.write_image(pos, rgb, format=out_format, mode=ImageMode.RGB)


def _float_to_rgba(pio_in, depth, transform, **kwargs):
    make_buf = lambda: np.empty((256, 256, 4), dtype=np.uint8)
    do_one = lambda buf, pos, pio_in, pio_out: _float_to_rgba_do_one(buf, pos, pio_in, pio_out, transform)
    _do_a_transform(pio_in, depth, make_buf, do_one, **kwargs)


def _float_to_rgba_do_one(buf, pos, pio_in, pio_out, transform):
    img = pio_in.read_image(pos, format='npy')
    if img is None:
        return

    mapped = transform(img.asarray())

    if mapped.ndim == 2:
        # TODO: allow real colormaps
        valid = np.isfinite(mapped)
        mapped[~valid] = 0
        mapped = np.clip(mapped * 255, 0, 255).astype(np.uint8)
        buf[...,:3] = mapped.reshape((256, 256, 1))
        buf[...,3] = 255 * valid
    else:
        valid = np.all(np.isfinite(mapped), axis=2)
        mapped[~valid] = 0
        mapped = np.clip(mapped * 255, 0, 255).astype(np.uint8)
        buf[...,:3] = mapped
        buf[...,3] = 255 * valid

    rgba = Image.from_array(buf)
    pio_out.write_image(pos, rgba, format='png', mode=ImageMode.RGBA)


# u8-to-RGB

def u8_to_rgb(pio, depth, **kwargs):
    make_buf = lambda: np.empty((256, 256, 3), dtype=np.uint8)
    _do_a_transform(pio, depth, make_buf, _u8_to_rgb_do_one, **kwargs)


def _u8_to_rgb_do_one(buf, pos, pio_in, pio_out):
    img = pio_in.read_image(pos, format='npy')
    if img is None:
        return

    buf[...] = img.asarray()[...,np.newaxis]
    rgb = Image.from_array(buf)
    pio_out.write_image(pos, rgb, format='jpg', mode=ImageMode.RGB)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Building up WWT imagery data sets.

This gets a little complex since the generation of a tiled image involves
several tasks that may or may not be implemented in several, swappable ways:
generating the tiled pixel data; positioning the image on the sky; filling in
metadata; and so on. We try to provide a framework that allows the
implementations of different tasks to be swapped out without getting too airy
and abstract.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
Builder
'''.split()

from wwt_data_formats.enums import DataSetType, ProjectionType
from wwt_data_formats.imageset import ImageSet
from wwt_data_formats.layers import ImageSetLayer, LayerContainerReader
from wwt_data_formats.place import Place

from .image import ImageLoader


class Builder(object):
    """
    State for some kind of imagery data set that's being assembled.
    """

    pio = None
    "A PyramidIO object representing the backing storage of the tiles and other image data."

    imgset = None
    """
    The WWT ImageSet data describing the image data and their positioning on the sky.

    Data URLs in this ImageSet should be populated as relative URLs.

    """
    place = None
    "The WWT Place data describing a default view of the image data."

    def __init__(self, pio):
        self.pio = pio

        self.imgset = ImageSet()
        self.imgset.name = 'Toasty'
        self.imgset.file_type = '.' + pio.get_default_format()
        self.imgset.url = pio.get_path_scheme() + self.imgset.file_type

        self.place = Place()
        self.place.foreground_image_set = self.imgset
        self.place.name = 'Toasty'


    def _check_no_wcs_yet(self):
        """
        The astrometric fields of ImageSet change their meaning depending on
        whether the image in question is tiled or not. Therefore, you'll get
        bogus results if change the tiling status *after* setting the
        astrometric information. This method should be called by other methods
        that control tiling in order to catch the issue if the user does things
        backwards.
        """
        if self.imgset.center_x != 0 or self.imgset.center_y != 0:
            raise Exception('order-of-operations error: you must apply WCS after applying tiling settings')


    def prepare_study_tiling(self, image):
        """
        Set up to tile the specified image as a WWT "study".

        Parameters
        ----------
        image : `toasty.image.Image`
            The image that will be tiled

        Returns
        -------
        tiling : `toasty.study.StudyTiling`
            The prepared tiling information

        Notes
        -----
        After calling this method, you should set up the WCS for the tiled
        imagery, using :meth:`default_tiled_study_astrometry` as a backstop if
        no real information is available. Then use :meth:`execute_study_tiling`
        to actually perform the tiling process.
        """

        from .study import StudyTiling

        tiling = StudyTiling(image.width, image.height)
        tiling.apply_to_imageset(self.imgset)
        return tiling


    def execute_study_tiling(self, image, tiling, **kwargs):
        """
        Tile the specified image as a WWT "study".

        Parameters
        ----------
        image : `toasty.image.Image`
            The image that will be tiled
        tiling : `toasty.study.StudyTiling`
            The prepared tiling information
        **kwargs
            Arguments relayed to :meth:`toasty.study.StudyTiling.tile_image`,
            such as ``cli_progress``.

        Returns
        -------
        *self*
        """

        tiling.tile_image(image, self.pio, **kwargs)
        return self


    def tile_base_as_study(self, image, **kwargs):
        """
        Tile an image assuming that it is in the appropriate format for WWT's
        "study" framework, namely that it uses a tangential (gnomonic)
        projection on the sky.

        Use of this method is somewhat discouraged since it both analyzes and
        performs the tiling all at once, which means that you can only correctly
        set (and validate) the WCS information *after* doing all the work of
        tiling. (Which in turn is because the proper way to apply WCS
        information to an imageset depends on the tiling parameters.) It is
        generally better to use :meth:`prepare_study_tiling` and
        :meth:`execute_study_tiling`, applying the WCS metadata in between, so
        that WCS errors can be caught and reported before doing the I/O.
        """

        from .study import tile_study_image

        self._check_no_wcs_yet()
        tiling = tile_study_image(image, self.pio, **kwargs)
        tiling.apply_to_imageset(self.imgset)

        return self


    def default_tiled_study_astrometry(self):
        self._check_no_wcs_yet()
        self.imgset.data_set_type = DataSetType.SKY
        self.imgset.base_degrees_per_tile = 1.0
        self.imgset.projection = ProjectionType.TAN
        self.place.zoom_level = 1.0
        return self


    def load_from_wwtl(self, cli_settings, wwtl_path, cli_progress=False):
        from contextlib import closing
        from io import BytesIO

        # Load WWTL and see if it matches expectations
        with closing(LayerContainerReader.from_file(wwtl_path)) as lc:
            if len(lc.layers) != 1:
                raise Exception('WWTL file must contain exactly one layer')

            layer = lc.layers[0]
            if not isinstance(layer, ImageSetLayer):
                raise Exception('WWTL file must contain an imageset layer')

            imgset = layer.image_set
            if imgset.projection != ProjectionType.SKY_IMAGE:
                raise Exception('WWTL imageset layer must have "SkyImage" projection type')

            # Looks OK. Read and parse the image.
            loader = ImageLoader.create_from_args(cli_settings)
            img_data = lc.read_layer_file(layer, layer.extension)
            img = loader.load_stream(BytesIO(img_data))

        # (Re-)initialize with the imageset info extracted from the WWTL.

        self.imgset = imgset
        self.place.foreground_image_set = self.imgset

        self.imgset.file_type = '.' + self.pio.get_default_format()
        self.imgset.url = self.pio.get_path_scheme() + self.imgset.file_type
        self.place.name = self.imgset.name

        # Transmogrify untiled image info to tiled image info. We reuse the
        # existing imageset as much as possible, but update the parameters that
        # change in the tiling process.

        wcs_keywords = self.imgset.wcs_headers_from_position()
        self.imgset.center_x = self.imgset.center_y = 0  # hack to satisfy _check_no_wcs_yet()
        self.tile_base_as_study(img, cli_progress=cli_progress)
        self.imgset.set_position_from_wcs(wcs_keywords, img.width, img.height, place=self.place)

        return img


    def toast_base(self, sampler, depth, is_planet=False, is_pano=False, **kwargs):
        from .toast import sample_layer

        self._check_no_wcs_yet()
        sample_layer(self.pio, sampler, depth, **kwargs)

        if is_planet:
            self.imgset.data_set_type = DataSetType.PLANET
        elif is_pano:
            self.imgset.data_set_type = DataSetType.PANORAMA
        else:
            self.imgset.data_set_type = DataSetType.SKY

        self.imgset.base_degrees_per_tile = 180
        self.imgset.projection = ProjectionType.TOAST
        self.imgset.tile_levels = depth
        self.place.zoom_level = 360

        return self


    def cascade(self, **kwargs):
        from .merge import averaging_merger, cascade_images
        cascade_images(self.pio, self.imgset.tile_levels, averaging_merger, **kwargs)
        return self


    def make_thumbnail_from_other(self, thumbnail_image):
        thumb = thumbnail_image.make_thumbnail_bitmap()
        with self.pio.open_metadata_for_write('thumb.jpg') as f:
            thumb.save(f, format='JPEG')
        self.imgset.thumbnail_url = 'thumb.jpg'

        return self


    def make_placeholder_thumbnail(self):
        import numpy as np
        from .image import Image

        arr = np.zeros((45, 96, 3), dtype=np.uint8)
        img = Image.from_array(arr)

        with self.pio.open_metadata_for_write('thumb.jpg') as f:
            img.aspil().save(f, format='JPEG')

        self.imgset.thumbnail_url = 'thumb.jpg'
        return self


    def apply_wcs_info(self, wcs, width, height):
        self.imgset.set_position_from_wcs(
            wcs.to_header(),
            width, height,
            place = self.place,
        )

        return self


    def apply_avm_info(self, avm, width, height):
        # So. The AVM standard discusses how parity should be expressed and how
        # it should be translated into WCS data, but in practice things are a
        # bit wonky: the AVM data that we've seen in the wild basically express
        # FITS-like (positive parity) WCS, while the actual associated image
        # data have a JPEG-like (negative parity) data layout. WCS can express
        # either parity so it would arguably be more correct for the generated
        # WCS to have negative parity. Based on the current state of knowledge,
        # I think the best option for now is to always flip the parity of the
        # WCS that pyavm hands us. We might need to change the heuristic or
        # allow the user to change the behavior.

        wcs = avm.to_wcs(target_shape=(width, height))
        from .image import _flip_wcs_parity
        wcs = _flip_wcs_parity(wcs, height)

        self.apply_wcs_info(wcs, width, height)

        if avm.Title:
            self.imgset.name = avm.Title

        if avm.Description:
            self.imgset.description = avm.Description

        if avm.Credit:
            self.imgset.credits = avm.Credit

        if avm.ReferenceURL:
            self.imgset.credits_url = avm.ReferenceURL

        return self


    def set_name(self, name):
        self.imgset.name = name
        self.place.name = name
        return self


    def create_wtml_folder(self):
        """
        Create a one-item :class:`wwt_data_formats.folder.Folder` object
        capturing this image.
        """
        from wwt_data_formats.folder import Folder

        self.place.name = self.imgset.name
        self.place.data_set_type = self.imgset.data_set_type
        self.place.thumbnail = self.imgset.thumbnail_url

        folder = Folder()
        folder.name = self.imgset.name

        # For all-sky/all-planet datasets, don't associate the imageset with a
        # particular Place. Otherwise, loading up the imageset causes the view
        # to zoom to a particular RA/Dec or lat/lon, likely 0,0. We might want
        # to make this manually configurable but this heuristic should Do The
        # Right Thing most times.
        if self.imgset.projection == ProjectionType.TOAST:
            folder.children = [self.imgset]
        else:
            folder.children = [self.place]

        return folder


    def write_index_rel_wtml(self):
        from wwt_data_formats import write_xml_doc

        folder = self.create_wtml_folder()

        with self.pio.open_metadata_for_write('index_rel.wtml') as f:
            write_xml_doc(folder.to_xml(), dest_stream=f, dest_wants_bytes=True)

        return self



# -*- mode: python; coding: utf-8 -*-
# Copyright 2013-2021 Chris Beaumont and the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
“Sampler” functions that fetch image data as a function of sky coordinates.

The Sampler Protocol
--------------------

A sampler is a callable object that obeys the following signature: ``func(lon,
lat) -> data``, where *lon* and *lat* are 2D numpy arrays of spherical
coordinates measured in radians, and the returned *data* array is a numpy
array of at least two dimensions whose first two axes have the same shape as
*lon* and *lat*. The *data* array gives the map values sampled at the
corresponding coordinates. Its additional dimensions can be used to encode
color information: one standard is for *data* to have a dtype of
``np.uint8`` and a shape of ``(ny, nx, 3)``, where the final axis samples
RGB colors.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
ChunkedPlateCarreeSampler
plate_carree_sampler
plate_carree_galactic_sampler
plate_carree_ecliptic_sampler
plate_carree_planet_sampler
healpix_fits_file_sampler
healpix_sampler
'''.split()

import numpy as np


def healpix_sampler(data, nest=False, coord='C', interpolation='nearest'):
    """Create a sampler for HEALPix image data.

    Parameters
    ----------
    data : array
        The HEALPix data
    nest : bool (default: False)
        Whether the data is ordered in the nested HEALPix style
    coord : 'C' | 'G'
        Whether the image is in Celestial (C) or Galactic (G) coordinates
    interpolation : 'nearest' | 'bilinear'
        What interpolation scheme to use.

        WARNING: bilinear uses healpy's get_interp_val, which seems prone to segfaults

    Returns
    -------
    A function that samples the HEALPix data; the call signature is
    ``vec2pix(lon, lat) -> data``, where the inputs and output are 2D arrays
    and *lon* and *lat* are in radians.

    """
    from healpy import ang2pix, get_interp_val, npix2nside
    from astropy.coordinates import Galactic, ICRS
    import astropy.units as u

    interp_opts = ['nearest', 'bilinear']
    if interpolation not in interp_opts:
        raise ValueError("Invalid interpolation %s. Must be one of %s" %
                         (interpolation, interp_opts))
    if coord.upper() not in 'CG':
        raise ValueError("Invalid coord %s. Must be 'C' or 'G'" % coord)

    galactic = coord.upper() == 'G'
    interp = interpolation == 'bilinear'
    nside = npix2nside(data.size)

    def vec2pix(l, b):
        if galactic:
            f = ICRS(l * u.rad, b * u.rad)
            g = f.transform_to(Galactic)
            l, b = g.l.rad, g.b.rad

        theta = np.pi / 2 - b
        phi = l

        if interp:
            return get_interp_val(data, theta, phi, nest=nest)

        return data[ang2pix(nside, theta, phi, nest=nest)]

    return vec2pix


def _find_healpix_extension_index(pth):
    """Find the first HEALPIX extension in a FITS file and return the extension
    number. Raises IndexError if none is found.

    """
    for i, hdu in enumerate(pth):
        if hdu.header.get('PIXTYPE') == 'HEALPIX':
            return i
    else:
        raise IndexError("No HEALPIX extensions found in %s" % pth.filename())


def healpix_fits_file_sampler(path, extension=None, interpolation='nearest'):
    """Create a sampler for HEALPix data read from a FITS file.

    Parameters
    ----------
    path : string
        The path to the FITS file.
    extension : integer or None (default: None)
        Which extension in the FITS file to read. If not specified, the first
        extension with PIXTYPE = "HEALPIX" will be used.
    interpolation : 'nearest' | 'bilinear'
        What interpolation scheme to use.

        WARNING: bilinear uses healpy's get_interp_val, which seems prone to segfaults

    Returns
    -------
    A function that samples the HEALPix image; the call signature is
    ``vec2pix(lon, lat) -> data``, where the inputs and output are 2D arrays
    and *lon* and *lat* are in radians.

    """
    from astropy.io import fits

    with fits.open(path) as f:
        if extension is None:
            extension = _find_healpix_extension_index(f)

        data, hdr = f[extension].data, f[extension].header

        # grab the first healpix parameter and convert to native endianness if
        # needed.
        data = data[data.dtype.names[0]]
        if data.dtype.byteorder not in '=|':
            data = data.byteswap().newbyteorder()

        nest = hdr.get('ORDERING') == 'NESTED'
        coord = hdr.get('COORDSYS', 'C')

    return healpix_sampler(data, nest, coord, interpolation)


def plate_carree_sampler(data):
    """Create a sampler function for all-sky data in a “plate carrée” projection.

    In this projection, the X and Y axes of the image correspond to the
    longitude and latitude spherical coordinates, respectively. Both axes map
    linearly, the X axis to the longitude range [2pi, 0] (i.e., longitude
    increases to the left), and the Y axis to the latitude range [pi/2,
    -pi/2]. Therefore the point with lat = lon = 0 corresponds to the image
    center and ``data[0,0]`` is the pixel touching lat = pi/2, lon=pi, one of
    a row adjacent to the North Pole. Typically the image is twice as wide as
    it is tall.

    Parameters
    ----------
    data : array-like, at least 2D
        The map to sample in plate carrée projection.

    Returns
    -------
    A function that samples the image; the call signature is
    ``vec2pix(lon, lat) -> data``, where the inputs and output are 2D arrays
    and *lon* and *lat* are in radians.

    """
    data = np.asarray(data)
    ny, nx = data.shape[:2]

    dx = nx / (2 * np.pi)  # pixels per radian in the X direction
    dy = ny / np.pi  # ditto, for the Y direction
    lon0 = np.pi - 0.5 / dx  # longitudes of the centers of the pixels with ix = 0
    lat0 = 0.5 * np.pi - 0.5 / dy  # latitudes of the centers of the pixels with iy = 0

    def vec2pix(lon, lat):
        lon = (lon + np.pi) % (2 * np.pi) - np.pi  # ensure in range [-pi, pi]
        ix = (lon0 - lon) * dx
        ix = np.round(ix).astype(int)
        ix = np.clip(ix, 0, nx - 1)

        iy = (lat0 - lat) * dy  # *assume* in range [-pi/2, pi/2]
        iy = np.round(iy).astype(int)
        iy = np.clip(iy, 0, ny - 1)

        return data[iy, ix]

    return vec2pix


def plate_carree_galactic_sampler(data):
    """
    Create a sampler function for all-sky data in a “plate carrée” projection
    using Galactic coordinates.

    Parameters
    ----------
    data : array-like, at least 2D
        The map to sample in plate carrée projection.

    Returns
    -------
    A function that samples the image. The call signature is
    ``sampler(lon, lat) -> data``, where the inputs and output are 2D arrays and
    *lon* and *lat* are in radians.

    """
    from astropy.coordinates import Galactic, ICRS
    import astropy.units as u

    data = np.asarray(data)
    ny, nx = data.shape[:2]

    dx = nx / (2 * np.pi)  # pixels per radian in the X direction
    dy = ny / np.pi  # ditto, for the Y direction
    lon0 = np.pi - 0.5 / dx  # longitudes of the centers of the pixels with ix = 0
    lat0 = 0.5 * np.pi - 0.5 / dy  # latitudes of the centers of the pixels with iy = 0

    def vec2pix(lon, lat):
        gal = ICRS(lon * u.rad, lat * u.rad).transform_to(Galactic)
        lon, lat = gal.l.rad, gal.b.rad

        lon = (lon + np.pi) % (2 * np.pi) - np.pi  # ensure in range [-pi, pi]
        ix = (lon0 - lon) * dx
        ix = np.round(ix).astype(int)
        ix = np.clip(ix, 0, nx - 1)

        iy = (lat0 - lat) * dy  # *assume* in range [-pi/2, pi/2]
        iy = np.round(iy).astype(int)
        iy = np.clip(iy, 0, ny - 1)

        return data[iy, ix]

    return vec2pix


def plate_carree_ecliptic_sampler(data):
    """
    Create a sampler function for all-sky data in a “plate carrée” projection
    using ecliptic coordinates.

    Parameters
    ----------
    data : array-like, at least 2D
        The map to sample in plate carrée projection.

    Returns
    -------
    A function that samples the image. The call signature is
    ``sampler(lon, lat) -> data``, where the inputs and output are 2D arrays and
    *lon* and *lat* are in radians.

    """
    from astropy.coordinates import BarycentricTrueEcliptic as Ecliptic, ICRS
    import astropy.units as u

    data = np.asarray(data)
    ny, nx = data.shape[:2]

    dx = nx / (2 * np.pi)  # pixels per radian in the X direction
    dy = ny / np.pi  # ditto, for the Y direction
    lon0 = np.pi - 0.5 / dx  # longitudes of the centers of the pixels with ix = 0
    lat0 = 0.5 * np.pi - 0.5 / dy  # latitudes of the centers of the pixels with iy = 0

    def vec2pix(lon, lat):
        ecl = ICRS(lon * u.rad, lat * u.rad).transform_to(Ecliptic)
        lon, lat = ecl.lon.rad, ecl.lat.rad
        lon = lon % (2 * np.pi) - np.pi  # ensure in range [-pi, pi]

        ix = (lon0 - lon) * dx
        ix = np.round(ix).astype(int)
        ix = np.clip(ix, 0, nx - 1)

        iy = (lat0 - lat) * dy  # *assume* in range [-pi/2, pi/2]
        iy = np.round(iy).astype(int)
        iy = np.clip(iy, 0, ny - 1)

        return data[iy, ix]

    return vec2pix


def plate_carree_planet_sampler(data):
    """
    Create a sampler function for planetary data in a “plate carrée” projection.

    This is the same as :func:`plate_carree_sampler`, except that the X axis
    is mirrored: longitude increases to the right. This is generally what is
    desired for planetary surface maps (looking at a sphere from the outside)
    instead of sky maps (looking at a sphere from the inside).

    Parameters
    ----------
    data : array-like, at least 2D
        The map to sample in plate carrée projection.

    Returns
    -------
    A function that samples the image; the call signature is
    ``vec2pix(lon, lat) -> data``, where the inputs and output are 2D arrays
    and *lon* and *lat* are in radians.

    """
    data = np.asarray(data)
    ny, nx = data.shape[:2]

    dx = nx / (2 * np.pi)  # pixels per radian in the X direction
    dy = ny / np.pi  # ditto, for the Y direction
    lon0 = -np.pi + 0.5 / dx  # longitudes of the centers of the pixels with ix = 0
    lat0 = 0.5 * np.pi - 0.5 / dy  # latitudes of the centers of the pixels with iy = 0

    def vec2pix(lon, lat):
        lon = (lon + np.pi) % (2 * np.pi) - np.pi  # ensure in range [-pi, pi]
        ix = (lon - lon0) * dx
        ix = np.round(ix).astype(int)
        ix = np.clip(ix, 0, nx - 1)

        iy = (lat0 - lat) * dy  # *assume* in range [-pi/2, pi/2]
        iy = np.round(iy).astype(int)
        iy = np.clip(iy, 0, ny - 1)

        return data[iy, ix]

    return vec2pix


class ChunkedPlateCarreeSampler(object):
    """
    Setup for TOAST sampling of a chunked plate-carree image.

    This only works with the ChunkedJPEG2000Reader image right now, but in
    principle we could extend to accept other chunked formats.

    Assume a typical image coordinate system, where (x,y) = (0,0) is the
    top-left edge of the image. The "global" coordinate system refers to the
    un-chunked image, which is probably too big to fit into memory.

    In the planetary plate carree projection, (global) X and Y measure latitude
    and longitude orthogonally. The left edge of the X=0 column has lon = -pi,
    while the right edge of the X=(W-1) column has lon = +pi. The top edge of
    the Y=0 row has lat = +pi/2, and the bottom edge of the Y=(H-1) row has lat
    = -pi/2.
    """

    def __init__(self, chunked_image, planetary=False):
        self._image = chunked_image

        assert planetary, 'XXX non-planetary plate carree not implemented'

        self.sx = 2 * np.pi / self._image.shape[1]  # radians per pixel, X direction
        self.sy = np.pi / self._image.shape[0]  # ditto, for the X direction

    @property
    def n_chunks(self):
        return self._image.n_chunks

    def _chunk_bounds(self, ichunk):
        cx, cy, cw, ch = self._image.chunk_spec(ichunk)

        # Note that the following are all intended to be coordinates at
        # pixel edges, in the appropriate direction, not pixel centers.
        lon_l = self.sx * cx - np.pi
        lon_r = self.sx * (cx + cw) - np.pi
        lat_u = 0.5 * np.pi - self.sy * cy
        lat_d = 0.5 * np.pi - self.sy * (cy + ch)  # note: lat_u > lat_d

        return lon_l, lon_r, lat_d, lat_u

    def filter(self, ichunk):
        """
        Get a TOAST tile-filter function for the specified chunk.

        Parameters
        ----------
        ichunk : :class:`int`
            The index of the chunk to query, between 0 and ``self.n_chunks - 1``
            (inclusive).

        Returns
        -------
        A callable object, ``filter(tile) -> bool``, suitable for use as a tile
        filter function.
        """
        chunk_lon_min1, chunk_lon_max1, chunk_lat_min1, chunk_lat_max1 = self._chunk_bounds(ichunk)

        def latlon_tile_filter(tile):
            """
            Return true if this tile might contain data from this chunk of the
            source image.
            """
            # Need to copy the outer arguments since we're going to modify them.
            chunk_lon_min = chunk_lon_min1
            chunk_lon_max = chunk_lon_max1
            chunk_lat_min = chunk_lat_min1
            chunk_lat_max = chunk_lat_max1

            corner_lonlats = np.asarray(tile.corners)  # shape (4, 2)

            # Latitudes are easy -- no wrapping.

            tile_lat_min = np.min(corner_lonlats[:,1])
            tile_lat_max = np.max(corner_lonlats[:,1])

            if chunk_lat_min > tile_lat_max:
                return False
            if chunk_lat_max < tile_lat_min:
                return False

            # Can't reject based on latitudes. Longitudes are trickier.
            #
            # Step 1: If we hit one of the poles, longitudes become
            # ~meaningless, so all we can really do (in our simplistic approach
            # here) is accept the tile. We compare to ~(pi - 1e-8) here to
            # account for inevitable roundoff. Note that if we don't apply the
            # lat-bounds filter first, every single chunk will accept tiles at
            # the poles, even if it's right on the equator.

            if tile_lat_max > 1.5707963 or tile_lat_min < -1.5707963:
                return True

            # Step 2: get good a coverage range for tile longitudes. Some tiles
            # span a full pi in longitude, and sometimes the "corners"
            # longitudes span all sorts of values from -2pi to 2pi. So just
            # shuffle them around until we get a self-consistent min/max.

            lons = corner_lonlats[:,0]
            keep_going = True

            while keep_going:
                keep_going = False

                imin = np.argmin(lons)
                tile_lon_min = lons[imin]
                tile_lon_max = np.max(lons)

                if tile_lon_max - tile_lon_min > np.pi:
                    keep_going = True
                    lons[imin] += 2 * np.pi

            # Step 3: Now, shuffle the chunk longitudes by +/-2pi so that they
            # line up with the tile longitude bounds. We know that we'll
            # ultimately be comparing the chunk_lon_min to tile_lon_max, and
            # chunk/max to tile/min, so we can treat those pairs individually.

            while chunk_lon_min - tile_lon_max > np.pi:
                chunk_lon_min -= 2 * np.pi

            while tile_lon_max - chunk_lon_min > np.pi:
                chunk_lon_min += 2 * np.pi

            while chunk_lon_max - tile_lon_min > np.pi:
                chunk_lon_max -= 2 * np.pi

            while tile_lon_min - chunk_lon_max > np.pi:
                chunk_lon_max += 2 * np.pi

            # Finally, we can reject this tile if it lies beyond the boundaries
            # of the chunk we're considering.

            if chunk_lon_min > tile_lon_max:
                return False
            if chunk_lon_max < tile_lon_min:
                return False

            # Well, we couldn't reject it, so we must accept:
            return True

        return latlon_tile_filter

    def sampler(self, ichunk):
        """
        Get a TOAST sampler function for the specified chunk.

        Parameters
        ----------
        ichunk : :class:`int`
            The index of the chunk to query, between 0 and ``self.n_chunks - 1``
            (inclusive).

        Returns
        -------
        A callable object, ``sampler(lon, lat) -> data``, suitable for use as a tile
        sampler function.
        """
        from .image import Image

        chunk_lon_min, chunk_lon_max, chunk_lat_min, chunk_lat_max = self._chunk_bounds(ichunk)
        data = self._image.chunk_data(ichunk)
        data_img = Image.from_array(data)
        buffer = data_img.mode.make_maskable_buffer(256, 256)
        biy, bix = np.indices((256, 256))

        ny, nx = data.shape[:2]
        dx = nx / (chunk_lon_max - chunk_lon_min)  # pixels per radian in the X direction
        dy = ny / (chunk_lat_max - chunk_lat_min)  # ditto, for the Y direction
        lon0 = chunk_lon_min + 0.5 / dx  # longitudes of the centers of the pixels with ix = 0
        lat0 = chunk_lat_max - 0.5 / dy  # latitudes of the centers of the pixels with iy = 0

        def plate_carree_planet_sampler(lon, lat):
            lon = (lon + np.pi) % (2 * np.pi) - np.pi  # ensure in range [-pi, pi]
            ix = (lon - lon0) * dx
            ix = np.round(ix).astype(int)
            ok = (ix >= 0) & (ix < nx)

            iy = (lat0 - lat) * dy  # *assume* in range [-pi/2, pi/2]
            iy = np.round(iy).astype(int)
            ok &= (iy >= 0) & (iy < ny)

            data_img.fill_into_maskable_buffer(buffer, iy[ok], ix[ok], biy[ok], bix[ok])
            return buffer.asarray()

        return plate_carree_planet_sampler



# -*- mode: python; coding: utf-8 -*-
# Copyright 2021 the AAS WorldWide Telescope project
# Licensed under the MIT License.

import os.path
import pytest

from . import HAS_AVM, test_path
from .. import cli


class TestAvm(object):
    def setup_method(self, method):
        from tempfile import mkdtemp
        self.work_dir = mkdtemp()

    def teardown_method(self, method):
        from shutil import rmtree
        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)

    @pytest.mark.skipif('not HAS_AVM')
    def test_check_cli_bad(self):
        args = [
            'check-avm',
            '--print',
            '--exitcode',
            test_path('badavm-type.png'),
        ]

        try:
            cli.entrypoint(args)
        except SystemExit as e:
            assert e.code == 1
        else:
            assert False, 'no error exit on bad AVM'

    @pytest.mark.skipif('not HAS_AVM')
    def test_check_cli_good(self):
        args = [
            'check-avm',
            '--print',
            test_path('geminiann11015a.jpg'),
        ]

        cli.entrypoint(args)
